#!/usr/bin/perl -w

# This script helps you keep your permissions in order (albeit you have to think about
# the order entries appear in your configuration file).
#
# (c) 2014 by Frederic Krueger / fkrueger-dev-updatecustomperms@holics.at
#
# Licensed under Creative Commons Attribution-ShareAlike 3.0 Unported
# There is no warranty of any kind, explicit or implied, for anything this software does or does not do.
# 
# Updates for this piece of software could be available under the following URL:
#   http://dev.techno.holics.at/update_custom_perms/
#
# This thing ;-) was created during trying to get rid of udev, mainly using devtmpfs instead.


## uses
use strict;
use v5.10;
use File::Basename;
use File::Find;
use Time::HiRes qw (time ualarm alarm);
use POSIX qw(strftime);
use Getopt::Std;


## settings
our $DEBUG = (0==1);
our $loglevel = 1;	# 0..6
our $itemsstep = 1000;	# in debug level 1, show update every 1000 files


## the following is used below in mayberesetfile() to decide what to do with exec bits
# on certain filenames; using 'file' has too big an overhead IMO. We could cache results, but.. naah :-)).
# regular expressions are being used, obviously.
#
# add your own as you see fit.
# - file extensions (think $fpath =~ /$archivers$/)
our $archivers  = "7z|ace|acf|bin|bz2|gz|ibak|img|iso|pak|pwsafe|rar|tar|tbz2|tgz|vpk|xz|zip";
our $compilers  = "cbproj|cc|cmake|cpp|d|filter|h|js|o|pbxproj|plistxcconfig|xcodeproj";
our $imagefiles = "bmp|dds|gif|jpeg|jpg|png|tga|xcf";
our $mediafiles = "mp3|wav|avi|flv|m4a|m4v|mkv|mov|mp4|ogg|ogm|ogv|vob|webm|wma|wmv|x264";
our $textfiles  = "cfg|conf|css|csv|doc|docx|htm|html|info|ini|log|lst|md5|pdf|pem|ppt|pptx|txt|xls|xlsx";
our $tempfiles  = "cache|dat|rdf|torrent|tmp|temp";
# - full filenames (think $fpath =~ /^$miscfiles$/)
our $miscfiles  = "id_rsa.*|aquota.*|ChangeLog.*|Makefile.*|README.*";
##

## defaults
our $regex_fileexts = qr /\.(?:$archivers|$compilers|$imagefiles|$mediafiles|$textfiles|$tempfiles)$/;
our $regex_fullfnames = qr /^$miscfiles$/;


## globals

our $CFG = "/etc/default/update_custom_perms";

our %config = ();
our %cfgline = ();		# current cfgline used for communicating via file::find::wanted

our %chmods = ();
our $chmodscnt = 0;
our %chowns = ();
our $chownscnt = 0;

our $curts = "not-set";		# used in logme() for now, gets rewritten every half a second by an alarm

our %cachepwgrids = ();		# to cache getpwuid/getgrgid data

# XXX optimization: create an alarm that rewrites $curts (the timestamp used for logging) every second; obviously the log entry can be off by a whole second using this.
local $SIG{'ALRM'} = sub { $curts = strftime ("%Y-%m-%d %H:%M:%S", localtime()); };
Time::HiRes::ualarm (1, 1000000);	# start rewriting $curts in 100ms, then every 500ms




## func
###
### DONT EDIT BELOW HERE
###

our ($PROG_NAME, $PROG_VERSION, $PROG_EMAIL, $PROG_URL) = ("update_custom_perms", "0.01", 'fkrueger-dev-updatecustomperms@holics.at', 'http://dev.techno.holics.at/update_custom_perms/');

## not a CHANGELOG:
## v0.01	first version created


## func
# this function is not called often; no optimization is needed here.
sub usage
{
  my ($msg) = @_;
  $msg = "" if (!defined($msg));

  print "\nusage: update_custom_perms.pl [-d <subdir-to-work-on>] [-u] [-h]\n";
  print "\n";
  print "   -f     provide file perm configuration (default: $CFG)\n";
  print "   -c     check file permissions for changes (according to the configuration above)\n";
  print "   -u     update changed file permissions automatically (warning: this can be dangerous, so make sure to check before!)\n";
  print "   -d     Provide subdir to work on (default: /)\n";
  print "   -l     set loglevel (0..6; default: 1)\n";
  print "   -h     this help\n";
  print "\n";
  print "   The configuration file is being created in $CFG .\n";
  print "   Edit as you see fit.\n";
  print "\n";
  print "This script allows you to check your mandatory acl filesystem for changes compared to the configured desired state.\n";
  print "\n";
  print "$PROG_NAME v$PROG_VERSION is licensed under Creative Commons Attribution-ShareAlike 3.0 Unported License.
There is no warranty of any kind, explicit or implied, for anything this software does or does not do.

The main page for this plugin can be found at: $PROG_URL

(c) 2011,2014 by Frederic Krueger / $PROG_EMAIL

";

  if ($msg ne "")
  {
    print "\n";
    print "Error: $msg\n";
    print "\n";
    print "\n";
  }
  exit (-1);
} # end sub usage


# caching getpwuid/getgrgid
sub cache_getpwgrids
{
  my ($type, $name) = @_;
  if (defined($cachepwgrids{$type}{$name}))
  {
    return ($cachepwgrids{$type}{$name});
  } # if grid/pwuid entry is in cache
  else	# 
  {
    $cachepwgrids {$type}{$name} = (($type eq "group") ? getgrgid($name) : getpwuid($name));
  } # if grid/pwuid entry is NOT incache
  return();	# never reached
}



# this function is called very often and thus needs more optimization.
sub getstat
{
  my ($fname) = @_;
#optimization:  logme (4, "+getstat(fname:$fname)");
  my %out = ( 'fname' => $fname, 'user' => -1, 'group' => -1, 'perms' => -1, 'perms_oct' => -1 );
  return ( { \%out } )	if ((!defined($fname)) or ($fname eq "") or (! -e $fname));

  my @t = stat($fname);
  if ($#t < 6)
  {
    logme (1, "Couldn't get stat on fname $fname")	if ($loglevel <= 1);
    return (\%out);
  }

  $out{'user'} = cache_getpwgrids  ("user", $t[4]);
  $out{'group'} = cache_getpwgrids ("group", $t[5]);
  $out{'perms'} = $t[2];
  $out{'perms_oct'} = sprintf ("%o", $out{'perms'});

  logme (4, "-getstat(fname:$fname) => user:" .(defined($out{'user'}) ? $out{'user'}:"undef"). ", group:" .(defined($out{'group'}) ? $out{'group'}:"undef"). ", perms:" .(defined($out{'perms'}) ? $out{'perms'}:"undef"). ", perms_oct:" .(defined($out{'perms_oct'}) ? $out{'perms_oct'}:"undef"))	      if ($loglevel <= 4);;
  return (\%out);
} # end sub getstat



# this function is called very often and thus needs more optimization.
sub mayberesetfile
{
  my ($fname, $refstartarr, $basedir) = @_;
  logme (4, "+mayberesetfile(fname:$fname)")      if ($loglevel <= 4);

  if (substr($fname, 0, length($basedir)) ne $basedir)
  {
    logme (3, "not in basedir for fname '$fname' (basedir:$basedir)")      if ($loglevel <= 3);
    return();
  }

  my %cfgline = %{ getcfgline($fname) };
  if ((! defined($cfgline{'user'})) or ($cfgline{'user'} eq "-1"))
  {
    logme (3, "got file-find result '$fname' without matching regexp-info. Skipping.")      if ($loglevel <= 3);
    return();
  }

  my %statinfo = %{ getstat($fname) };
  if ((! defined($statinfo{'user'})) or ($statinfo{'user'} eq "-1"))
  {
    logme (3, "got stat-info result '$fname' with unknown information. Skipping.")      if ($loglevel <= 3);
    return();
  }


  ### chown part
  if (($statinfo{'user'} ne $cfgline{'user'}) or ($statinfo{'group'} ne $cfgline{'group'}))
  {
    logme (2, "creating chown-entry for fname '$fname': (stat vs. cfg: u: ".$statinfo{'user'}."<->".$cfgline{'user'}. ", g: " .$statinfo{'group'}. "<->" .$cfgline{'group'}. ")")      if ($loglevel <= 2);
    if ((++$chownscnt % $itemsstep) == 0) { logme (1, "" .dirname($fname). ": Created $chownscnt chown-entries so far"); }
    my $usergrpcombo = $cfgline{'user'}.":".$cfgline{'group'};
    if (!defined($chowns{$usergrpcombo}))  { $chowns{$usergrpcombo} = []; }
    push @{$chowns{$usergrpcombo}}, $fname;
    logme (4, "chowns-$usergrpcombo now has " .($#{$chowns{$usergrpcombo}}+1). " entries")      if ($loglevel <= 4);
  }

  my $workperm = -1;	my $workperm_oct = -1;
  if (-d $fname)
  {
    $workperm = $cfgline{'perms-d'};
    $workperm_oct = $cfgline{'perms-d_oct'};
  }
  elsif (-e $fname)
  {
    $workperm = $cfgline{'perms-f'};
    $workperm_oct = $cfgline{'perms-f_oct'};
  }
  else
  {
    logme (3, "skipping weirdling '$fname'")      if ($loglevel <= 3);
  }
  logme (4, "setting workperm to $workperm (oct:$workperm_oct)")      if ($loglevel <= 4);

  ### chmod part
  if (($workperm >= 0) and ($workperm_oct >= 0))
  {
    logme (4, "stat<->work perms are " .sprintf("%o", $statinfo{'perms'} & 07777). "<->" .sprintf("%o", $workperm & 07777))      if ($loglevel <= 4);
    if (($statinfo{'perms'} & 07777) ne ($workperm & 07777))
    {
      my $specialparts = $statinfo{'perms'} & 0777000;
      my $execparts    = $statinfo{'perms'} & 0000111;
      my $otherparts   = $statinfo{'perms'} & 0000666;

      # imported from way above!
      # remove exec-bits on the following filenames automatically
      if (($fname =~ $regex_fileexts) or ($fname =~ $regex_fullfnames))
      {
        if (! -d $fname)
        {
          logme (5, "removing useless exec-bits on fname '$fname'")      if ($loglevel <= 5);
	  $execparts = 0;
# XXX if you want to keep the "useless" pre-existing exec-bits, use the line below
#          $workperm = $workperm & 0777666; $workperm_oct = sprintf ("%o", $workperm);
        }
      }

      logme (4, "found chmod-fname '$fname': (stat<->cfg perms are ".$statinfo{'perms_oct'}. "<->$workperm_oct)")      if ($loglevel <= 4);

      my $newperm = $workperm | $execparts | $specialparts;
      my $newperm_oct = sprintf ("%o", $newperm);

      logme (4, "stat exec: $execparts (oct:" .sprintf("%o", $execparts). "), other: $otherparts (oct:" .sprintf("%o", $otherparts). "), special: $specialparts (oct:" .sprintf("%o", $specialparts). ")")      if ($loglevel <= 4);

      # only add chmod-entry if we have an actual change in permissions
      if ($newperm != $statinfo{'perms'})
      {
        logme (2, "creating chmod-entry for fname $fname from " .$statinfo{'perms_oct'}. " to $newperm_oct")      if ($loglevel <= 2);
        if ((++$chmodscnt % $itemsstep) == 0) { logme (1, "" .dirname($fname). ": Created $chmodscnt chmod-entries so far"); }
        if (!defined($chmods{$newperm}))  { $chmods{$newperm} = []; }
	push @{$chmods{$newperm}}, $fname;
	logme (4, "chmods-$newperm now has " .($#{$chmods{$newperm}}+1). " entries")      if ($loglevel <= 4);
      }
    } # end if file-perms (&0777) are different
  } # end if not a weirdling
} # end sub mayberesetfile




# this function is called very often and thus needs more optimization.
sub logme
{
  my ($level, $msg) = @_;
  $level = 0	if (!defined($level));
  return ()	if ((!$DEBUG) and ($level > $loglevel));

  # XXX optimization: $curts is being rewritten via SIGVTALRM (using Time::HiRes::alarm to rewrite it every half a second)
  printf ("%s %s %s\n", $curts, "."x$level, (defined($msg)?$msg:""));
#  my $out = sprintf ("%s %s %s\n", $curts, "."x$level, (defined($msg)?$msg:""));
#  print "$out";
  return ();
} # end sub logme


# this function is not called often; no optimization is needed here.
sub glob2regex
{
  my ($glob) = @_;

  my $regex = $glob;
  $regex =~ s/\./\\./g;
  $regex =~ s/\*/.*/g;
  $regex =~ s/\?/./g;

  return ($regex);
} # end sub glob2regex



# this function is called very often and thus needs more optimization.
sub getcfgline
{
  my ($fname) = @_;
  return ({})	if ((! defined($fname)) or ($fname eq "") or (! -e $fname));

  logme (5, "+getcfgline(fname:$fname)")      if ($loglevel <= 5);

  my %matchstrings = ();
  my ($key, $curmatch);
  foreach $key (keys %{$config{'orig'}})
  {
    logme (6, "checking '$fname' =~ '$key': regex '" .$config{'orig'}{$key}{'fpath_regex'}. "'")      if ($loglevel <= 6);
    if ($fname =~ $config{'orig'}{$key}{'fpath_regex'})	# find exact match of fname-string matching the cfg entry, using qr'd regex entry
    {
      $curmatch = $1;
      logme (5, "got a match for key '$key' with length " .(length($curmatch)). ": '$curmatch'")      if ($loglevel <= 5);
      $matchstrings { $curmatch } = $key;
    }
  } # end foreach keyline

  my $matchfname;
  foreach $matchfname (sort { length($a) <=> length($b) } keys %matchstrings) # get longest match
  {
    if ($matchstrings{$matchfname} ne "")
    {
      logme (5, "got longest match (length:".length($matchfname). ") for fname '$fname', key '$matchstrings{$matchfname}'")      if ($loglevel <= 5);
      return (\%{$config{'orig'}{ $matchstrings{$matchfname} }});
    }
  } # end foreach matchfname

  logme (5, "no match for fname '$fname'")      if ($loglevel <= 5);
  return ({});
} # end sub getcfgline





# the following parts are not called often; no optimization is needed here.

## args

our ($opt_c, $opt_u, $opt_f, $opt_b, $opt_d, $opt_l, $opt_h);
our $basedir = "/";			# by default: /
our $whattodo = "nonespecified";	# by default: only check


if ($#ARGV < 0)
{
  &usage;
}
else
{
  getopts('cuf:b:dl:h');
}


if (defined($opt_h))
{
  usage();
}


## and the actual arg parsing
if (defined($opt_c))
{
  $whattodo = "check";
}

if (defined($opt_u))
{
  $whattodo = "update";
}
if ($whattodo eq "nonespecified")
{
  usage ("You must choose a mode to run, either -c for check or -u for update permissions.");
}



if ((defined($opt_f)) and ($opt_f ne "") and (-e $opt_f))
{
  $CFG = $opt_f;
}
else
{
  $CFG = "/etc/default/update_custom_perms";
  if (! -e "$CFG")
  {
    if (-d "$CFG")
    {
      open (C, ">$CFG");
      print C "##filepath to dev device(s)\towner user\towner group\tchmod perms for files (-exec bits)\tchmod perms for dirs
##a little hint: the longest match will be used/applied, ie. the entry for /tmp/foobar.txt is being applied before /tmp .
##also, everything is applied recursively; use ie. /tmp for dirs, and /tmp/foo/* for the files in /tmp/foo/.
##think: bash/dos-style globbing.
#/dev/dri/*      root    video   0660
#/dev/media*     root    video   0660
#/dev/pcmC0D*    root    audio   0660
#/dev/snd/*      root    audio   0660
#/dev/video*     root    video   0660
#*/aquota.*      root    users   0600    2700
";
      close (C);
    }
    else
    {
      usage ("Please create the directory for the configuration file $CFG .");
    }
  } # end if cfg doesnt exist
} # end if using default cfg place


 
if ((defined($opt_b)) and ($opt_b ne ""))
{
  $basedir = $opt_b;
}
if (! -d $basedir)
{
  usage ("Specified directory must exist !");
}

if (defined($opt_d))
{
  $DEBUG = (0==0);
}

if (defined($opt_l))
{
  if (($opt_l >= 0) and ($opt_l <= 6))
  {
    $loglevel = int($opt_l);
  }
  else
  {
    usage ("Log level must be between 0 and 6 !"); 
  }
}

##
## main
##

my $mysettings = "" .(($whattodo ne "nonespecified") ? "mode:$whattodo ":""). ((defined($basedir)) ? "basedir:$basedir ":""). "loglevel:$loglevel ". ((defined($opt_d)) ? "debug:$DEBUG ":"");
logme (1, "Started in $whattodo mode (${mysettings}cfg:$CFG)")      if ($loglevel <= 1);

# to save time lateron, prepare most of the regexp
if (-e $CFG)
{
  logme (2, "Reading from cfgfile $CFG")      if ($loglevel <= 2);
  my $cntid = 1;		# start ids with 1
  open (CFG, "<$CFG") || die "!!! Couldn't open cfgfile $CFG: $!\n";
  while (my $inp = <CFG>)
  {
    chomp ($inp);
    next	if ($inp =~ /^\s{0,}#/);
  
    my @info = split /\s+/, $inp;
    if ($#info < 3)
    {
      logme (3, "found $#info infoline: '$inp'")      if ($loglevel <= 3);
      next;
    }
  
    if (rindex($info[0], "/") != length($info[0])) { $info[0] =~ s/\/+$//g; }
    if (defined($config{$info[0]}))
    {
      print "!!! found dupe fpath '$info[0]' in configuration: '$inp'\n";
      next;
    }
  
    my $curid = $cntid++;
    my $fpath = $info[0];
    my $fpath_regex = glob2regex($fpath);
    my $perms_f = oct($info[3]);
    my $perms_d = oct($info[4]);
    logme (3, "found fpath-info '$fpath' (re-version: $fpath_regex) has user:$info[1], group:$info[2], perms-f:$info[3], perms-d:$info[4]")      if ($loglevel <= 3);
    $config{'orig'}{$fpath_regex} = { 'fpath' => $fpath, 'fpath_regex' => qr /^($fpath_regex)(.*)/, 'user' => $info[1], 'group' => $info[2], 'perms-f' => $perms_f, 'perms-f_oct' => $info[3], 'perms-d' => $perms_d, 'perms-d_oct' => $info[4], 'id' => $curid };
  
    my @tmparr = ();
  
    ####
    if (index($fpath, "*") >= 0)
    {
      logme (4, "found glob '$fpath' with *")      if ($loglevel <= 4);
      push @tmparr, glob($fpath);
    }
    else
    {
      if (-d $fpath)
      {
        logme (4, "found dir '$fpath'")      if ($loglevel <= 4);
        push @tmparr, $fpath;
      } # end if found dir
      elsif (-e $fpath)
      {
        logme (4, "found file '$fpath'")      if ($loglevel <= 4);
        push @tmparr, $fpath;
      } # end if found file
      else
      {
        logme (4, "found weirdling '$fpath'.. Skipping.")      if ($loglevel <= 4);
      }
    } # end if a direct filename
  
    logme (4, "now adding found targets to lib-path")      if ($loglevel <= 4);
    # now add all the found filepaths in @tmparr to our little paths' library ;)
    my $fn;
    foreach $fn (@tmparr)
    {
      if ($fn =~ /$basedir/)
      {
        logme (5, "adding lib-path '$fn'")      if ($loglevel <= 5);
        $config{'lib'}{$fn} = { %{$config{'orig'}{$fpath_regex}} };
      }
      else
      {
        logme (5, "not adding lib-path '$fn': doesn't apply with basedir $basedir")      if ($loglevel <= 5);
      }
    }
    ####
  
  }
  close (CFG);
} # end if configuration exists


if (($whattodo eq "check") or ($whattodo eq "update"))
{
  logme (1, "")      if ($loglevel <= 1);
  
  my @tmp = keys %{$config{'lib'}};
  my @sortedlist = sort { $a cmp $b } @tmp;
  
  our @startarr = @sortedlist;
  
  if ($#startarr < 0)
  {
    logme (2, "No matching configuration entries found. Exitting.")      if ($loglevel <= 2);
    exit();
  }
  else
  {
    logme (1, "Found " .($#startarr+1). " matching configuration entries.")      if ($loglevel <= 1);
  }
  
  
  logme (1, "Now running find on all matching configuration entries:")      if ($loglevel <= 1);
  
  # the following parts create/show/process massive amounts of data; optimization may be needed here.
  find(\&wanted, @startarr);
  sub wanted
  {
    mayberesetfile ($File::Find::name, \@startarr, $basedir);
  } # end sub wanted
} # end if mode is check or update
else
{
  logme (1, "Mode is neither 'check' nor 'update'; exitting.")      if ($loglevel <= 1);
  exit (-1);
}



# XXX on the rc:
# return rc != 0 for mode 'check', if things to change were found
# return rc == 0 for mode 'update', if all changes could be done successfully

my $rc = ($whattodo eq "update") ? (0==0) : (0==1);

logme (1, "")      if ($loglevel <= 1);
my @permlist = keys %chmods;
my @usrgrplist = keys %chowns;
logme (1, "Collected " .($#permlist+1). " chmods and " .($#usrgrplist+1). " chowns")      if ($loglevel <= 1);

if ($#permlist >= 0)
{
  logme (1, "")      if ($loglevel <= 1);
  logme (1, "Doing " .(($whattodo eq "update")?"nothing for ":""). ($#permlist+1). " permissions for chmods ...")      if ($loglevel <= 1);
  my $curperm;
  foreach $curperm (keys %chmods)
  {
    if ($whattodo eq "check") { $rc = (0==1); }
    logme (1, "  for perms $curperm")      if ($loglevel <= 1);
    logme (2, "  " .join (",", @{$chmods{$curperm}}))      if ($loglevel <= 2);
    if ($whattodo eq "update") { if ($rc) { $rc = chmod ($curperm, @{$chmods{$curperm}}); } }
  } # end foreach aref [fname, perms] entry
}

if ($#usrgrplist >= 0)
{
  logme (1, "")      if ($loglevel <= 1);
  logme (1, "Doing " .(($whattodo eq "update")?"nothing for ":""). ($#usrgrplist+1). " combos for chowns ...")      if ($loglevel <= 1);
  my $curusrgrpcombo;
  foreach $curusrgrpcombo (keys %chowns)
  {
    my ($curuser, $curgroup) = ($curusrgrpcombo =~ /^([^:]+):(.*)$/);
    my $curuid = getpwnam($curuser);
    my $curgid = getgrnam($curgroup);
    logme (1, "  for usrgrpcombo $curusrgrpcombo (user:$curuser $curuid, group:$curgroup $curgid)")      if ($loglevel <= 1);
    logme (2, "  " .join (",", @{$chowns{$curusrgrpcombo}}))      if ($loglevel <= 2);
    if ($whattodo eq "update") { if ($rc) { $rc = chown ($curuid, $curgid, @{$chowns{$curusrgrpcombo}}); } }
  } # end foreach aref [fname, perms] entry
}


logme (1, "#fin")      if ($loglevel <= 1);
exit ($rc);



